/*
 * $Id: ClassVisitor.java 8 2008-02-21 16:01:23Z bellingard $
 *
 * (C) Copyright 2005 Diomidis Spinellis
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

package gr.spinellis.ckjm;

import org.apache.bcel.Constants;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.Type;

import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.TreeSet;

/**
 * Visit a class updating its Chidamber-Kemerer metrics.
 *
 * @see ClassMetrics
 * @version $Revision: 1.19 $
 * @author <a href="http://www.spinellis.gr">Diomidis Spinellis</a>
 */
public class ClassVisitor extends org.apache.bcel.classfile.EmptyVisitor {
    
    /** The class being visited. */
    private JavaClass visitedClass;
    
    /** The class's constant pool. */
    private ConstantPoolGen cp;
    
    /** The class's fully qualified name. */
    private String myClassName;
    
    /** The container where metrics for all classes are stored. */
    private ClassMetricsContainer cmap;
    
    /** The emtrics for the class being visited. */
    private ClassMetrics cm;
    
    /** Classes encountered.
     * Its cardinality is used for calculating the CBO.
     */
    private HashSet efferentCoupledClasses = new HashSet();
    
    /** Methods encountered.
     * Its cardinality is used for calculating the RFC.
     */
    private HashSet responseSet = new HashSet();
    
    /** Use of fields in methods.
     * Its contents are used for calculating the LCOM.
     * We use a Tree rather than a Hash to calculate the
     * intersection in O(n) instead of O(n*n).
     */
    private ArrayList mi = new ArrayList();

    /**
     * Constructor
     * @param jc java class
     * @param classMap container
     */
    public ClassVisitor(JavaClass jc, ClassMetricsContainer classMap) {
        visitedClass = jc;
        cp = new ConstantPoolGen(visitedClass.getConstantPool());
        cmap = classMap;
        myClassName = jc.getClassName();
        cm = cmap.getMetrics(myClassName);
    }

    /**
     * @return the class's metrics container
     */
    public ClassMetrics getMetrics() {
        return cm;
    }

    /**
     * visit java class
     */
    public void start() {
        visitJavaClass(visitedClass);
    }

    /** 
     * Calculate the class's metrics based on its elements.
     * @see org.apache.bcel.classfile.Visitor#visitJavaClass(org.apache.bcel.classfile.JavaClass)
     */
    public void visitJavaClass(JavaClass jc) {
        String super_name = jc.getSuperclassName();
        String package_name = jc.getPackageName();

        cm.setVisited();
        if (jc.isPublic()) {
            cm.setPublic();
        }
        ClassMetrics pm = cmap.getMetrics(super_name);

        pm.incNoc();
        /*
         * JRE's Object has Object as its parent.
         * Skip it or else we'll get into infinite recursion
         * when calculating DIT.
         */
        if (!jc.getClassName().equals("java.lang.Object")) {
            cm.setParent(pm);
        }
        registerCoupling(super_name);

        String ifs[] = jc.getInterfaceNames();
        /* Measuring decision: couple interfaces */
        for (int i = 0; i < ifs.length; i++) {
            registerCoupling(ifs[i]);
        }

        Field[] fields = jc.getFields();
        for (int i = 0; i < fields.length; i++) {
            fields[i].accept(this);
        }

        Method[] methods = jc.getMethods();
        for (int i = 0; i < methods.length; i++) {
            methods[i].accept(this);
        }
    }

    /**
     * Add a given class to the classes we are coupled to
     * @param className the class' name
     */
    public void registerCoupling(String className) {
        /* Measuring decision: don't couple to Java SDK */
        if ((MetricsFilter.isJdkIncluded() || !ClassMetrics.isJdkClass(className)) && !myClassName.equals(className)) {

            cmap.getMetrics(className).addAfferentCoupling(myClassName);
        }
    }

    /**
     * @param className the class' name
     */
    public void registerCouplingCbo(String className) {
        /* Measuring decision: don't couple to Java SDK */

        if ((MetricsFilter.isJdkIncluded() || !ClassMetrics.isJdkClass(className)) && !myClassName.equals(className)) {
            efferentCoupledClasses.add(className);

        }
    }

    /**
     * Add the type's class to the classes we are coupled to
     * @param t type
     */
    public void registerCoupling(Type t) {
        registerCoupling(className(t));
    }

    /**
     * Add a given class to the classes we are coupled to
     * @param className the class' name
     * @param fieldName class to add
     */
    void registerFieldAccess(String className, String fieldName) {
        registerCoupling(className);
        if (className.equals(myClassName)) {
            ((TreeSet) mi.get(mi.size() - 1)).add(fieldName);
        }
    }

    /**
     * Add a given method to our response set
     * @param className class' name
     * @param methodName method's name
     * @param args arguments
     */
    void registerMethodInvocation(String className, String methodName, Type[] args) {
        registerCoupling(className);
        /* Measuring decision: calls to JDK methods are included in the RFC calculation */
        incRFC(className, methodName, args);
    }

    /** Called when a field access is encountered.
     * @see org.apache.bcel.classfile.Visitor#visitField(org.apache.bcel.classfile.Field)
     */
    public void visitField(Field field) {
        registerCoupling(field.getType());
    }

    /**
     * Called when encountering a method that should be included in the class's RFC.
     * @param className class' name
     * @param methodName method's name
     * @param arguments arguments
     */
    private void incRFC(String className, String methodName, Type[] arguments) {
        String argumentList = Arrays.asList(arguments).toString();
        // remove [ ] chars from begin and end
        String args = argumentList.substring(1, argumentList.length() - 1);
        String signature = className + "." + methodName + "(" + args + ")";
        responseSet.add(signature);
    }

    /** 
     * Called when a method invocation is encountered.
     * @see org.apache.bcel.classfile.Visitor#visitMethod(org.apache.bcel.classfile.Method)
     */
    public void visitMethod(Method method) {
        MethodGen mg = new MethodGen(method, visitedClass.getClassName(), cp);

        Type result_type = mg.getReturnType();
        Type[] argTypes = mg.getArgumentTypes();

        registerCoupling(mg.getReturnType());
        for (int i = 0; i < argTypes.length; i++) {
            registerCoupling(argTypes[i]);
        }

        String[] exceptions = mg.getExceptions();
        for (int i = 0; i < exceptions.length; i++) {
            registerCoupling(exceptions[i]);
        }

        /* Measuring decision: A class's own methods contribute to its RFC */
        incRFC(myClassName, method.getName(), argTypes);

        cm.incWmc();

        if (Modifier.isPublic(method.getModifiers())) {
            cm.incNpm();

        }
        mi.add(new TreeSet());
        MethodVisitor factory = new MethodVisitor(mg, this);
        factory.start();
    }

    /**
     * Return a class name associated with a type.
     * @param t type
     * @return class name
     */
    static String className(Type t) {
        String ts = t.toString();
        String result = ts;
        if (t.getType() <= Constants.T_VOID) {
            result = "java.PRIMITIVE";
        } else if (t instanceof ArrayType) {
            ArrayType at = (ArrayType) t;
            result = className(at.getBasicType());
        }
        return result;
    }
    
    /** Do final accounting at the end of the visit. */
    public void end() {

        cm.setCbo(efferentCoupledClasses.size());
        cm.setRfc(responseSet.size());
        /*
         * Calculate LCOM  as |P| - |Q| if |P| - |Q| > 0 or 0 otherwise
         * where
         * P = set of all empty set intersections
         * Q = set of all nonempty set intersections
         */
        int lcom = 0;
        for (int i = 0; i < mi.size(); i++) {
            for (int j = i + 1; j < mi.size(); j++) {
                /* A shallow unknown-type copy is enough */
                TreeSet intersection = (TreeSet) ((TreeSet) mi.get(i)).clone();
                intersection.retainAll((TreeSet) mi.get(j));
                if (intersection.size() == 0) {
                    lcom++;
                } else {
                    lcom--;
                }
            }
        }
        cm.setLcom(lcom > 0 ? lcom : 0);
    }

    /**
     * @return class name
     */
    public String getMyClassName() {
        return myClassName;
    }

}
