/*
 * $Id: MethodVisitor.java 8 2008-02-21 16:01:23Z bellingard $
 *
 * (C) Copyright 2005 Diomidis Spinellis
 *
 * Permission to use, copy, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

package gr.spinellis.ckjm;

import org.apache.bcel.Constants;
import org.apache.bcel.generic.ArrayInstruction;
import org.apache.bcel.generic.CHECKCAST;
import org.apache.bcel.generic.CodeExceptionGen;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.ConstantPushInstruction;
import org.apache.bcel.generic.EmptyVisitor;
import org.apache.bcel.generic.FieldInstruction;
import org.apache.bcel.generic.GETFIELD;
import org.apache.bcel.generic.GETSTATIC;
import org.apache.bcel.generic.INSTANCEOF;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionConstants;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.LocalVariableInstruction;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.Type;

/**
 * Visit a method calculating the class's Chidamber-Kemerer metrics.
 * A helper class for ClassVisitor.
 *
 * @see ClassVisitor
 * @version $Revision: 1.8 $
 * @author <a href="http://www.spinellis.gr">Diomidis Spinellis</a>
 */
class MethodVisitor extends EmptyVisitor {

    /** Method generation template. */
    private MethodGen mg;

    /** The class's constant pool. */
    private ConstantPoolGen cp;

    /** The visitor of the class the method visitor is in. */
    private ClassVisitor cv;

    /** The metrics of the class the method visitor is in. */
    private ClassMetrics cm;

    /**
     * Constructor.
     * @param m method
     * @param c visitor
     */
    MethodVisitor(MethodGen m, ClassVisitor c) {
        mg = m;
        cv = c;
        cp = mg.getConstantPool();
        cm = cv.getMetrics();
    }
    
    /**  
     * Start the method's visit.
     */
    public void start() {
        if (!mg.isAbstract() && !mg.isNative()) {
            for (InstructionHandle ih = mg.getInstructionList().getStart(); ih != null; ih = ih.getNext()) {
                Instruction i = ih.getInstruction();
                if (!visitInstruction(i)) {
                    i.accept(this);
                }
            }
            updateExceptionHandlers();
        }
    }

    /**
     * Visit a single instruction.
     * @param i instruction
     * @return result of the visit
     */
    private boolean visitInstruction(Instruction i) {
        short opcode = i.getOpcode();
        return ((InstructionConstants.INSTRUCTIONS[opcode] != null) && !(i instanceof ConstantPushInstruction) && !(i instanceof ReturnInstruction));
    }

    /**
     * Local variable use.
     * @param i local variable
     */
    public void visitLocalVariableInstruction(LocalVariableInstruction i) {
        if (i.getOpcode() != Constants.IINC) {
            cv.registerCoupling(i.getType(cp));
        }
    }

    /**
     * Array use.
     * @param i instructions array
     */
    public void visitArrayInstruction(ArrayInstruction i) {
        cv.registerCoupling(i.getType(cp));
    }

    /**
     * Field access.
     * @param i field
     */
    public void visitFieldInstruction(FieldInstruction i) {
        cv.registerFieldAccess(i.getClassName(cp), i.getFieldName(cp));
        cv.registerCoupling(i.getFieldType(cp));
        registerCouplingCboField(i);

    }

    /**
     * Method invocation.
     * @param i invocation
     */
    public void visitInvokeInstruction(InvokeInstruction i) {
        Type[] argTypes = i.getArgumentTypes(cp);
        for (int j = 0; j < argTypes.length; j++) {
            cv.registerCoupling(argTypes[j]);
        }
        cv.registerCoupling(i.getReturnType(cp));

        if (!(i instanceof INVOKESPECIAL)) {
            cv.registerCouplingCbo(i.getClassName(cp));
        }
        /* Measuring decision: measure overloaded methods separately */
        cv.registerMethodInvocation(i.getClassName(cp), i.getMethodName(cp), argTypes);

    }

    /**
     * Visit an instanceof instruction.
     * @param i instanceof instruction
     */
    public void visitINSTANCEOF(INSTANCEOF i) {
        cv.registerCoupling(i.getType(cp));
    }

    /**
     * Visit checklast instruction.
     * @param i checklast instruction
     */
    public void visitCHECKCAST(CHECKCAST i) {
        cv.registerCoupling(i.getType(cp));
    }

    /** Visit return instruction.
     * @param i return instruction 
     */
    public void visitReturnInstruction(ReturnInstruction i) {
        cv.registerCoupling(i.getType(cp));
    }

    /**
     * Visit the method's exception handlers.
     */
    private void updateExceptionHandlers() {
        CodeExceptionGen[] handlers = mg.getExceptionHandlers();

        /* Measuring decision: couple exceptions */
        for (int i = 0; i < handlers.length; i++) {
            Type t = handlers[i].getCatchType();
            if (t != null) {
                cv.registerCoupling(t);
            }
        }
    }

    /**
     * @param i field instruction
     */
    private void registerCouplingCboField(FieldInstruction i) {

        if ((i instanceof GETFIELD) || (i instanceof GETSTATIC)) {

            if (!(cv.getMyClassName().equals(i.getClassName(cp)))) { //fields of current class visite                  
                cv.registerCouplingCbo(ClassVisitor.className(i.getFieldType(cp)));
            }
        }
    }

}
