package com.airfrance.squaleide.util;

import java.io.File;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;

import com.airfrance.squaleide.SqualePlugin;
import com.airfrance.squaleide.preferences.SqualePreferencePage;
import com.airfrance.squaleide.util.messages.PluginMessages;

/**
 * Mthodes utilitaires pour la gestion des marqueurs
 */
public class MarkerHelpers {

    /** Le nom du rpertoire jsp rajout au nom des composants JSP */
    public static final String JSP_PACKAGE = "{jsp}";

    /**
     * Cre un log
     * @param project le projet
     * @param desc la description du message
     * @return le log
     * @throws CoreException si erreur
     */
    public static IMarker createLogMarker(IProject project, String desc) throws CoreException {
        return createMarker(project, IMarker.SEVERITY_ERROR, IMarker.PRIORITY_HIGH, desc, null, null, 0);
    }

    /**
     * Cre une erreur
     * @param project le projet
     * @param desc la description du message
     * @param path le rpertoire contenant la ressource
     * @param name le nom du fichier
     * @param line la ligne o se trouve l'erreur
     * @return l'erreur
     * @throws CoreException si erreur
     */
    public static IMarker createErrorMarker(IProject project, String desc, String path, String name, int line) throws CoreException {
        return createMarker(project, IMarker.SEVERITY_ERROR, IMarker.PRIORITY_HIGH, desc, path, name, line);
    }

    /**
     * Cre un warning
     * @param project le projet
     * @param desc la description du message
     * @param path le rpertoire contenant la ressource
     * @param name le nom du fichier
     * @param line la ligne o se trouve le warning
     * @return le warning
     * @throws CoreException si erreur
     */
    public static IMarker createWarningMarker(IProject project, String desc, String path, String name, int line) throws CoreException {
        return createMarker(project, IMarker.SEVERITY_WARNING, IMarker.PRIORITY_NORMAL, desc, path, name, line);
    }

    /**
     * Cre une info
     * @param project le projet
     * @param desc la description du message
     * @param path le rpertoire contenant la ressource
     * @param name le nom du fichier
     * @param line la ligne o se trouve l'info
     * @return l'info
     * @throws CoreException si erreur
     */
    public static IMarker createInfoMarker(IProject project, String desc, String path, String name, int line) throws CoreException {
        return createMarker(project, IMarker.SEVERITY_INFO, IMarker.PRIORITY_LOW, desc, path, name, line);
    }

    /**
     * Cre un marqueur pour une transgression
     * exemple de transgression : 
     * <code>transgression description="Les scriptlets doivent ?tre plac?s dans des Tag Libraries ou des d?clarations JSP" 
     * inFolder="vobs/squale/src/squaleWeb/WebContent/jsp/canvas/" 
     * location="27" 
     * resource="application_menu.jsp" 
     * severity="warning"</code>
     * @param project le projet
     * @param severity la svrit de la transgression
     * @param priority la priorit de la transgression
     * @param desc la description de la transgression
     * @param path le chemin du rpertoire dans lequel se trouve la ressource associe  la transgression
     * @param name le nom du fichier dans lequel se trouve la transgression
     * @param location le numro de ligne de la transgression
     * @return le marqueur associ  la transgression
     * @throws CoreException si erreur
     */
    private static IMarker createMarker(IProject project, int severity, int priority, String desc, String path, String name, int location) throws CoreException {
        if (path.length() > 0 && !path.endsWith("/")) {
            path += "/";
        }
        IResource res = findResource(project, path, name);
        boolean found = true;
        // Message  rajouter dans le cas o la ressource sera non cliquable
        String msgIfResNotFound = "";
        if (null == res) {
            // On lie la tche au projet
            res = project.getWorkspace().getRoot();
            found = false;
        }
        IMarker marker = res.createMarker(IMarker.TASK);
        marker.setAttribute(IMarker.SEVERITY, severity);
        marker.setAttribute(IMarker.PRIORITY, priority);
        // On ne met jamais 0 car eclipse ne connat pas cette ligne et dans le cas o
        // le fichier est dj ouvert, il ne se repositionne pas  la ligne 1.
        if (0 == location) {
            location = 1;
            // De plus, on vrifie que la transgression ne se porte pas sur un mthode
            // car sinon on rajoute son nom pour localiser
            // le composant impliqu dans la transgression ( dfaut du numro de ligne)
            int firstBracket = name.lastIndexOf("(");
            if (firstBracket > 0) {
                // On ajoute le nom  la description
                desc += " (mthode : " + getMethodName(name, firstBracket) + ")";
            }
        }
        if(res.getType() != IResource.FILE) {
            // On met un message d'information pour informer que le ressource est non cliquable
            msgIfResNotFound = PluginMessages.getString("marker.ressource.not.found");
        }
        marker.setAttribute(IMarker.MESSAGE, "Squale - " + desc + " " + msgIfResNotFound);
        marker.setAttribute(IMarker.LINE_NUMBER, location);
        marker.setAttribute("SQUALE", "true");
        if (!found) {
            // Si la ressource n'existe pas, on indique la localisation indique dans le fichier XML
            marker.setAttribute(IMarker.LOCATION, path + name);
        }
        return marker;
    }

    /**
     * @param name le nom absolu de la mthode
     * @param bracketId l'index de la "("
     * @return le nom de la mthode sans les packages et classes
     */
    private static String getMethodName(String name, int bracketId) {
        String result = name;
        int lastDot = name.lastIndexOf(".", bracketId);
        if (lastDot > 0) {
            result = name.substring(lastDot + 1);
        }
        return result;
    }

    /**
     * PRECONDITION : path et fileName ne doivent pas tre nuls.
     * @param project le projet
     * @param path le chemin du rpertorie contenant le fichier
     * @param fileName le nom du fichier ou le nom d'un composant SQUALE
     * @return le fichier ou le rpertoire correspondant aux paramtres. Le projet sinon.
     */
    private static IResource findResource(IProject project, String path, String fileName) {
        // Null si la ressource n'est pas trouv dans le workspace
        IResource res = null;
        // On construit le chemin
        String pathfile = path + fileName;
        // On recherche d'abord un fichier
        IFile file = project.getFile(pathfile);
        if (file.exists()) {
            res = file;
        } else {
            // On recherche un fichier ou un rpertoire en construisant le chemin
            // par rapport au nom du fileName
            File projectLocation = project.getLocation().toFile();
            // On va parcourir l'arborescence du projet juqu' obtenir le fichier
            // correspondant au nom du composant
            File newRes = findFileInWorkspace(buildFileName(path, fileName), projectLocation);
            // On rcupre la ressource si elle a t trouve
            if (null != newRes) {
                // Il nous faut le chemin relatif au projet
                pathfile = PluginUtils.replaceFirst(projectLocation.getAbsolutePath().replace('\\', '/'), newRes.getAbsolutePath().replace('\\', '/'));
                if (newRes.isDirectory()) {
                    res = project.getFolder(pathfile);
                } else {
                    res = project.getFile(pathfile);
                }
            }
        }
        return res;
    }

    /**
     * @param path le chemin vers le fichier
     * @param fileName le nom du fichier ou le nom absolu du composant
     * @return le chemin (suppos)  partir du premier package (ou rpertoire) vers le fichier
     */
    private static String buildFileName(String path, String fileName) {
        String result = "";
        // Si le chemin vers le fichier n'est pas vide alors on va essayer de rcuprer
        // le fichier d'aprs les prfrences car il s'agit peut-tre d'un chemin en local
        if (path.length() > 0) {
            String absolutePath = (path + fileName).replace('\\', '/');
            // On rcupre la liste des chemins vers les sources
            String srcList = SqualePlugin.getDefault().getPreferenceStore().getString(SqualePreferencePage.P_LIST);
            String[] srcs = PluginUtils.parseString(srcList, SqualePreferencePage.SRC_SEPARATOR);
            for (int i = 0; i < srcs.length; i++) {
                String srcTrans = srcs[i].replace('\\', '/');
                if (!srcTrans.endsWith("/")) {
                    srcTrans += "/";
                }
                if (absolutePath.startsWith(srcTrans)) {
                    result = getFileNameWithoutExtension(PluginUtils.replaceFirst(srcTrans, absolutePath));
                }
            }
        } else {
            result = clear(fileName).replace('.', '/');
        }
        return result;
    }

    /**
     * Nettoie le nom du composant. 
     * Par exemple pour les composants JSP, le package {jsp} et rajout au nom
     * il faut donc le supprimer car il ne correspondra pas  un rpertoire existant
     * dans le workspace du projet.
     * @param componentName le nom du composant
     * @return le nom du composant nettoy.
     */
    private static String clear(String componentName) {
        String clean = PluginUtils.replaceFirst(JSP_PACKAGE, componentName);
        // On ne garde pas le nom de la mthode
        int firstBracketId = clean.lastIndexOf("(");
        if (firstBracketId > 0) {
            int lastDotId = clean.lastIndexOf(".", firstBracketId);
            if (lastDotId > 0) {
                clean = clean.substring(0, lastDotId);
            }
        }
        //  ni le nom d'une classe interne
        int firstDollard = clean.indexOf("$");
        if (firstDollard > 0) {
            clean = clean.substring(0, firstDollard);
        }
        return clean;
    }

    /**
     * @param pComponentName le nom du composant
     * @param pDirectory le rpertoire de rechecherche
     * @return le fichier de nom <code>componentName</code> contenu dans l'arborescence de
     * <code>container</code>
     */
    private static File findFileInWorkspace(String pComponentName, File pDirectory) {
        File result = null;
        // On rcupre tous les fichiers et dossiers du rpertoire pDirectory
        File[] files = pDirectory.listFiles(new ExtensionsFileFilter());
        if (files != null) {
            // On parcours rcursivement les rpertoires
            // jusqu' trouver le fichier dont le chemin absolu fini par componentPath
            for (int i = 0; i < files.length && result == null; i++) {
                File file = files[i];
                // On compare le nom du fichier sans l'ventuelle extension du fichier
                // car le nom du composant ne prsente pas l'extension.
                if (getFileNameWithoutExtension(file.getAbsolutePath().replace('\\', '/')).endsWith(pComponentName)) {
                    // On considre qu'il s'agit du fichier cherch
                    // Si plusieurs fichiers ont le mme nom dans le workspace, le premier
                    // trouv, sera le premier renvoy...
                    return file;
                } else if (file.isDirectory()) {
                    // On continue  chercher
                    result = findFileInWorkspace(pComponentName, file);
                }
            }
        }
        // Si aucun fichier ne correspond au critre, on renvoie null.
        return result;
    }

    /**
     * @param pFilename le nom du fichier
     * @return le nom du fichier sans son extension
     */
    private static String getFileNameWithoutExtension(String pFilename) {
        String result = pFilename;
        int lastDot = pFilename.lastIndexOf(".");
        if (lastDot > 0) {
            result = pFilename.substring(0, lastDot);
        }
        return result.replace('\\', '/');
    }

    /**
     * Supprime les marqueurs SQUALE
     * @param project le projet
     * @throws CoreException si erreur
     */
    public static void removeMarker(IProject project) throws CoreException {
        IMarker[] markers = project.getWorkspace().getRoot().findMarkers(null, true, IResource.DEPTH_INFINITE);
        if (markers != null) {
            int i, len = markers.length;
            for (i = 0; i < len; i++) {
                if (markers[i].getAttribute("SQUALE") != null) {
                    markers[i].delete();
                }
            }
        }
    }
}
